/**
 * @file servicelog_notify.c
 * @brief Program for registering/manipulating servicelog notification tools
 *
 * Copyright (C) 2005, 2008  IBM
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301,
 * USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#define _GNU_SOURCE
#include <getopt.h>
#include <servicelog-1/servicelog.h>
#include "config.h"

#define ACTION_TOOMANY		-1
#define ACTION_UNSPECIFIED	0
#define ACTION_ADD		1
#define ACTION_LIST		2
#define ACTION_REMOVE		3

#define ARG_LIST	"amrqi:t:E:R:c:M:l:h"

static char *cmd;

static struct option long_options[] = {
	{"add",		    no_argument,        NULL, 'a'},
	{"remove",	    no_argument,        NULL, 'r'},
	{"list",	    no_argument,        NULL, 'l'},
	{"match",	    required_argument,  NULL, 'm'},
	{"type",	    required_argument,  NULL, 't'},
	{"command",	    required_argument,	NULL, 'c'},
	{"method",	    required_argument,	NULL, 'M'},
	{"id",	            required_argument,	NULL, 'i'},
	{"help",	    no_argument,        NULL, 'h'},
	{0,0,0,0}
};

/**
 * print_usage
 * @brief Print the usage message
 */
static void
print_usage() 
{
	printf("Usage: %s {--add | --remove | --list} (flags}\n",
	       cmd);
	printf("  Add Flags:\n");
	printf("    --command=\"<cmd>\"  command to be execed when notified\n");
	printf("    --type=\"<type>\"    EVENT or REPAIR\n");
	printf("    --match=<query_string>\n");
	printf("    --method={num_stdin|num_arg|text_stdin|pairs_stdin}\n");
	printf("  Remove Flags: (one of --id or --command must be "
	       "specified)\n");
	printf("    --id=<id>          unique key of registered tool to "
	       "remove\n");
	printf("    --command=\"<cmd>\"  command name of registered tool to "
	       "remove\n");
	printf("  List Flags:\n");
	printf("    --command=\"<cmd>\"  find tools registered with <cmd>\n");
	printf("    --id=<id>          key of registered tool to find\n");
	printf("  Other Flags:\n");
	printf("    --help             print this help text and exit\n");

	return;
}

/**
 * valid_type_arg
 * @brief Validate an arg for a specified option
 *
 * @param type string argument to the type option
 * @return converted value if it is valid, -1 otherwise
 */
static int
valid_type_arg(char *type)
{
	if (strncmp(type, "EVENT", 5) == 0)
		return SL_NOTIFY_EVENTS;
	else if (strncmp(type, "REPAIR", 6) == 0)
		return SL_NOTIFY_REPAIRS;

	return -1;
}

/**
 * valid_method_arg
 * @brief validate an arg as either "num_stdin", "num_arg" or "text_stdin"
 *
 * @param arg argument to validate
 * @return converted value if it is valid, -1 otherwise
 */
static int
valid_method_arg(const char *arg) {
	if (strncmp(arg, "num_stdin", 9) == 0)
		return SL_METHOD_NUM_VIA_STDIN;
	else if (strncmp(arg, "num_arg", 7) == 0)
		return SL_METHOD_NUM_VIA_CMD_LINE;
	else if (strncmp(arg, "text_stdin", 10) == 0)
		return SL_METHOD_PRETTY_VIA_STDIN;
	else if (strncmp(arg, "pairs_stdin", 11) == 0)
		return SL_METHOD_SIMPLE_VIA_STDIN;

	return -1;
}

/**
 * main
 * @brief Parse command line args and execute diagnostics
 *
 * @param argc the number of command-line arguments
 * @param argv array of command-line arguments
 * @return exit status: 0 for normal exit, 1 for usage error, >1 for other error
 */
int
main(int argc, char *argv[]) 
{
	int option_index, rc, action=ACTION_UNSPECIFIED;
	int flag_type=0, flag_cmd=0, flag_match=0, flag_id=0, flag_method=0;
	int other_flag=0, type=0, method=0;
	uint64_t id=0;
	char *command=NULL, *match=NULL, query[256];
	struct servicelog *servlog;
	struct sl_notify *notify, *current;

	cmd = argv[0];

	if (argc <= 1) {
		print_usage();
		exit(0);
	}
		
	memset(&servlog, 0, sizeof(servlog));

	for (;;) {
		option_index = 0;
		rc = getopt_long(argc, argv, ARG_LIST, long_options,
				 &option_index);

		if (rc == -1)
			break;

		switch (rc) {
		case 'a':
			if (action != ACTION_UNSPECIFIED)
				action = ACTION_TOOMANY;
			if (action != ACTION_TOOMANY)
				action = ACTION_ADD;
			break;
		case 'r':
			if (action != ACTION_UNSPECIFIED)
				action = ACTION_TOOMANY;
			if (action != ACTION_TOOMANY)
				action = ACTION_REMOVE;
			break;
		case 'l':
			if (action != ACTION_UNSPECIFIED)
				action = ACTION_TOOMANY;
			if (action != ACTION_TOOMANY)
				action = ACTION_LIST;
			break;
		case 'i':	/* event ID */
			flag_id = 1;
			id = atoi(optarg);
			break;
		case 't':
			type = valid_type_arg(optarg);
			if (type == -1) {
				print_usage(argv[0]);
				exit(-1);
			}
			other_flag = 1;
			flag_type = 1;
			break;
		case 'c':
			command = optarg;
			flag_cmd = 1;
			break;
		case 'm':
			match = optarg;
			flag_match = 1;
			break;
		case 'M':
			method = valid_method_arg(optarg);
			if (method == -1) {
				print_usage(argv[0]);
				exit(-1);
			}
			other_flag = 1;
			flag_method = 1;
			break;
		case 'h':	/* help */
			print_usage(argv[0]);
			exit(0);
			break;
		case '?':
			print_usage(argv[0]);
			exit(1);
			break;
		default:
			printf("huh?\n");
		}
	}

	/* Command-line validation */
	if (action == ACTION_UNSPECIFIED) {
		fprintf(stderr, "One of --add, --remove, or --list "
			"is required.\n\n");
		print_usage(argv[0]);
		exit(-1);
	}

	if (action == ACTION_TOOMANY) {
		fprintf(stderr, "Only one of the --add, --remove, or "
			"--list options may be specified.\n\n");
		print_usage(argv[0]);
		exit(-1);
	}

	rc = servicelog_open(&servlog, 0);
	if (rc != 0) {
		fprintf(stderr, "%s\n", strerror(rc));
		return 2;
	}

	switch (action) {

	case ACTION_ADD:
		/* additional command line validation */
		if (flag_id) {
			fprintf(stderr, "The --id flag may not be used with "
				"the --add option.\n\n");
			print_usage(argv[0]);
			rc = -1;
			goto err_out;
		}
		if (command == NULL) {
			fprintf(stderr, "The --command flag must be specified "
				"with the --add option.\n\n");
			print_usage(argv[0]);
			rc = -1;
			goto err_out;
		}

		/* set up an sl_notify struct and call servicelog_notify_add */
		notify = (struct sl_notify *)malloc(sizeof(struct sl_notify));
		memset(notify, 0, sizeof(struct sl_notify));
		notify->notify = type;
		notify->method = method;
		if (match) {
			notify->match = malloc(strlen(match) + 1);
			strcpy(notify->match, match);
		}
		notify->command = malloc(strlen(command) + 1);
		strcpy(notify->command, command);

		rc = servicelog_notify_log(servlog, notify, &id);
		servicelog_notify_free(notify);
		if (rc == 0) {
			printf("Registration successful (id: %llu)\n", id);
		}
		else {
			fprintf(stderr, "%s\n", servicelog_error(servlog));
			rc = 2;
			goto err_out;
		}
		break;

	case ACTION_LIST:
		/* additional command line validation */
		if ((command != NULL) && (flag_id)) {
			fprintf(stderr, "Only one of the --command or --id "
				"flags may be specified with the --list "
				"option.\n\n");
			print_usage(argv[0]);
			rc = -1;
			goto err_out;
		}
		if (other_flag) {
			fprintf(stderr, "Neither the --type nor --method flags "
				"may be used with the --list option.\n\n");
			print_usage(argv[0]);
			rc = -1;
			goto err_out;
		}

		/* query the database */
		if (flag_id) {
			servicelog_notify_get(servlog, id, &notify);
			if (notify == NULL) {
				fprintf(stderr, "Could not find a registered "
					"notification tool with the specified "
					"id (%llu).\n", id);
				rc = 1;
				goto err_out;
			}
		}
		else if (command) {
			snprintf(query, 256, "command = '%s'", command);
			servicelog_notify_query(servlog, query, &notify);
			if (notify == NULL) {
				fprintf(stderr, "Could not find a registered "
					"notification tool with the specified "
					"command ('%s').\n", command);
				rc = 1;
				goto err_out;
			}
		}
		else {
			servicelog_notify_query(servlog, "id>0", &notify);
			if (notify == NULL) {
				fprintf(stderr, "There are no registered "
					"notification tools.\n");
				rc = 1;
				goto err_out;
			}
		}

		/* diplay the notification tools */
		servicelog_notify_print(stdout, notify, 2);
		servicelog_notify_free(notify);
		break;

	case ACTION_REMOVE:
		/* additional command line validation */
		if ((command == NULL) && (!flag_id)) {
			fprintf(stderr, "At least one of the --command or --id "
				"flags must be specified with the --remove "
				"option.\n\n");
			print_usage(argv[0]);
			rc = -1;
			goto err_out;
		}

		/* find the registered notification tool to be modified */
		if (flag_id) {
			servicelog_notify_get(servlog, id, &notify);
			if (notify == NULL) {
				fprintf(stderr, "Could not find a registered "
					"notification tool with the specified "
					"id (%llu).\n", id);
				rc = -1;
				goto err_out;
			}
		}
		else {		/* search for the notify tool by command name */
			snprintf(query, 256, "command = '%s'", command);
			servicelog_notify_query(servlog, query, &notify);
			if (notify == NULL) {
				fprintf(stderr, "Could not find a registered "
					"notification tool with the specified "
					"command ('%s').\n", command);
				rc = -1;
				goto err_out;
			}
			if (notify->next != NULL) {
				fprintf(stderr, "Multiple matches for the "
					"specified command:\n\n");
				current = notify;
				while (current) {
					servicelog_notify_print(stdout,
								current, 1);
					current = current->next;
				}
				servicelog_notify_free(notify);
				rc = 1;
				goto err_out;
			}
			id = notify->id;
		}

		servicelog_notify_delete(servlog, id);
		break;

	default:
		fprintf(stderr, "Internal error; unknown action %d\n", action);
		rc = -1;
		goto err_out;
	}

err_out:
	servicelog_close(servlog);

	return rc;
}
